--***************************************************************************
-- Copyright (c) 1994 - Ohio Board of Regents and
--                    - University of Cincinnati, Cincinnati, OH.
--                    - Sidhartha Mohanty,
--                    - Department of Electrical & Computer Engineering,
--                    - Sept, 1994.
--***************************************************************************
-- 
-- Title:               Design library for Queueing Model. Version 0.0.
-- 
-- File:                [SUITE: vhdl/project/queue] fork.vhdl
-- Created:             Sat Sept 17, 11:12:25 1994
-- Author:              Sidhartha Mohanty <smohanty@thor.ece.uc.edu>
-- 
-- Keywords:
-- LRM_version:         IEEE Std. P1076-1992/C
-- 
-- Objective:
--
-- Usage:
-- Analyze after queue packages and random package.
--
-- Algorithm:               
-- (None)
-- 
-- References:
-- 
-- Result:
-- 
-- History:
-- 
-- Comments:
-- 
--***************************************************************************

library queue;
use queue.queue_pkg.all;
use queue.rng.all;
use std.textio.all;

entity fork is
 
 generic ( out_control   : distrib     := unif;
           number_out    : integer     := 1;
	   name 	 : string      := "F1";
	   seed		 : integer     := 1;
	   generation    : frequency   := (0.0, 1.0);
	   debug_control : debug       := none;
           file_name     : string      := "";
	   report_freq   : time        := 10 fs);

 port ( in_arc : inout arc bus; 
	out_arc : inout arc_array (1 to number_out) bus);

end;

architecture behavior of fork is

  signal id  : integer := 1;
  file outfile : text is out file_name;
  
begin
  
initialize : process

   variable integer_line : line;
   variable integer_id   : integer := 1;

  begin

    in_arc <= null;
    out_arc <= null;

    assert not std.textio.endfile(std.textio.input)
	report "(fork.vhdl) Number of entities in Queueing Model exceed number of integer ids. Increase the number."
	severity failure;

    readline(std.textio.input, integer_line);

    assert integer_line'length >= 0
	report "(fork.vhdl) Missing integer in identifiers file."
	severity failure;

    read(integer_line, integer_id);
    id <= integer_id;

    wait for 0 fs;

    wait;

end process;

fork : process 

 variable j : integer := 1;
 variable L : line;
 variable unf   : Uniform := InitUniform(seed, generation.mean, generation.var);
 variable erlan : Erlang  := InitErlang(seed, generation.mean, generation.var);
 variable nex   : NegExp  := InitNegExp(seed, generation.mean);
 variable rnd   : Real    := 1.0;
 variable number_tokens_passed : integer := 0;
 variable mesg : line;

 begin  
   
  in_arc <= null;
--   for I in 1 to number_out loop
     out_arc <= null;
--   end loop;

  loop

    wait on in_arc until (in_arc.handshake = sending);
     number_tokens_passed := number_tokens_passed + 1;

     if (debug_control = handshaking or 
  	 debug_control = summary) then
       print_handshake_message ("Fork unit: ", name," received handshake signal: ", 
  			         in_arc.handshake,
                                " from component: ", in_arc.id,
  			        " at Time: ", now);
     end if;
     
     if (out_control = unif) then
          assert false
	    report "Routing of tokens follow uniform distribution"
	    severity note;
	GenRnd(unf);
        rnd := unf.rnd;
	assert rnd <= 1.0
	  report "Random Number Generated wrong, must be <= 1.0"
	  severity error; 
	for I in 1 to number_out loop
	  if ((real(I-1)/real(number_out) < rnd) and (rnd <= real(I)/real(number_out))) then
	     j := I;
	  end if;
	end loop; 
        out_arc(j) <= in_arc;
        wait for 0 fs;
        wait on out_arc until (out_arc(j).handshake = accepted);
          if (debug_control = handshaking or 
  	      debug_control = summary) then
              print_handshake_message ("Fork unit: ", name, " received handshake signal: ", 
  		  		        out_arc(j).handshake,
                                       " from component: ", out_arc(j).id,
  				       " at Time: ", now);
          end if;
        out_arc(j) <= null;
        wait for 0 fs;

     elsif (out_control = erl) then
	 assert false
	    report "Routing of tokens follow erlang distribution"
	    severity note;
	GenRnd(erlan);
        rnd := erlan.rnd;
	assert rnd <= 1.0
	  report "Random Number Generated wrong, must be <= 1.0"
	  severity error; 
	for I in 1 to number_out loop
	  if ((real(I-1)/real(number_out) < rnd) and (rnd <= real(I)/real(number_out))) then
	     j := I;
	  end if;
	end loop; 
        out_arc(j) <= in_arc;
        wait for 0 fs;
        wait on out_arc until (out_arc(j).handshake = accepted);
          if (debug_control = handshaking or 
  	      debug_control = summary) then
          print_handshake_message ("Fork unit: ", name, 
				   " received handshake signal: ", 
  				    out_arc(j).handshake,
                                   " from component: ", out_arc(j).id,
  				   " at Time: ", now);
          end if;
        out_arc(j) <= null;
        wait for 0 fs;

     elsif (out_control = negex) then
	 assert false
	    report "Routing of tokens follow negative exponential distribution"
	    severity note;
	GenRnd(nex);
        rnd := nex.rnd;
	assert rnd <= 1.0
	  report "Random Number Generated wrong, must be <= 1.0"
	  severity error; 
	for I in 1 to number_out loop
	  if ((real(I-1)/real(number_out) < rnd) and (rnd <= real(I)/real(number_out))) then
	     j := I;
	  end if;
	end loop; 
        out_arc(j) <= in_arc;
        wait for 0 fs;
        wait on out_arc until (out_arc(j).handshake = accepted);
          if (debug_control = handshaking or 
  	      debug_control = summary) then
          print_handshake_message ("Fork unit: ", name, " received handshake signal: ", 
  				    out_arc(j).handshake,
                                   " from component: ", out_arc(j).id,
  				   " at Time: ", now);
          end if;
        out_arc(j) <= null;
        wait for 0 fs;

     elsif (out_control = fixed) then
	 assert false
	    report "Routing of tokens follow fixed distribution"
	    severity note;
	if j > number_out then j := 1;
	end if;
	while j <= number_out loop
         out_arc(j) <= in_arc;
         wait for 0 fs;
         wait on out_arc until (out_arc(j).handshake = accepted);
           if (debug_control = handshaking or 
  	       debug_control = summary) then
           print_handshake_message ("Fork unit: ", name, 
	 			    " received handshake signal: ", 
  	 			     out_arc(j).handshake,
                                    " from component: ", out_arc(j).id,
  				    " at Time: ", now);
           end if;
         out_arc(j) <= null;
         wait for 0 fs;
	 j := j + 1;
	 exit;
	end loop;

     elsif (out_control = replicate) then

      write (L, string'("Fork of Replicate type, Input sent to all the"));
      write (L, string'(" output arcs at the same time"));
      writeline(std.textio.output, L);

      for I in 1 to number_out loop
  	if (debug_control = trace_tokens or 
  	    debug_control = summary) then
  	print_joinfork_message ("Fork unit: ", name, " passing tokens to ",
                                 "output arc(i)", " at Time: ", now);
  	end if;
        out_arc(i) <= in_arc;
	wait for 0 fs;
        wait on out_arc until (out_arc(i).handshake = accepted);
  	if (debug_control = handshaking or 
  	    debug_control = summary) then
  	print_handshake_message ("Fork unit: ", name," received handshake signal: ", 
  				 out_arc(i).handshake,
                                  " from component: ", out_arc(i).id,
  				 " at Time: ", now);
  	end if;
        out_arc(i) <= null;
        wait for 0 fs;
      end loop;

     else assert false
	    report "Out_control wrongly specified"
	    severity error;
  
    end if;

    write(mesg, string'("Total number of tokens passed through Fork unit "));
    write(mesg, name);
    write(mesg, string'(" at time: "));
    write(mesg, now);
    write(mesg, string'(" = "));
    write(mesg, number_tokens_passed);
    writeline(outfile, mesg);
    in_arc <= (id, accepted, FALSE, now);
    wait for 0 fs;
    if (debug_control = handshaking or 
  	debug_control = summary) then
    print_handshake_message ("Fork unit: ", name, " sent handshake signal: ",
                              in_arc.handshake, " to component: ",
  		   	      in_arc.id, " at Time: ", now);
    end if;
    in_arc <= null;
    wait for 0 fs;
  end loop;
 end process; 

end behavior;
