--***************************************************************************
-- Copyright (c) 1994 - Ohio Board of Regents and
--                    - University of Cincinnati, Cincinnati, OH.
--                    - Sidhartha Mohanty,
--                    - Department of Electrical & Computer Engineering,
--                    - Sept, 1994.
--***************************************************************************
-- 
-- Title:               Design library for Queueing Model. Version 0.0.
-- 
-- File:                [SUITE: vhdl/project/queue] join.vhdl
-- Created:             Fri Sept 16, 15:30:15 1994
-- Author:              Sidhartha Mohanty <smohanty@thor.ece.uc.edu>
-- 
-- Keywords:
-- LRM_version:         IEEE Std. P1076-1992/C
-- 
-- Objective:
--
-- Usage:
-- Analyze after queue packages and random package.
--
-- Algorithm:               
-- (None)
-- 
-- References:
-- 
-- Result:
-- 
-- History:
-- 
-- Comments:
-- 
--***************************************************************************

library queue;
use queue.queue_pkg.all;
use std.textio.all;

entity join is
 
 generic ( number_in     : integer := 1;
	   name 	 : string  := "J1";
	   debug_control : debug   := none;
           file_name     : string  := "";
	   report_freq   : time    := 10 fs);

 port ( in_arc : inout arc_array (1 to number_in) bus; 
	out_arc : inout arc bus);

end;

architecture behavior of join is

  signal join_ready : bit := '0';
  signal id 	    : integer := 1;
  file outfile : text is out file_name;
  
begin

initialize : process

   variable integer_line : line;
   variable integer_id   : integer := 1;

  begin

    out_arc <= null;

    assert not std.textio.endfile(std.textio.input)
	report "(join.vhdl) Number of entities in Queueing Model exceed number of integer ids. Increase the number."
	severity failure;

    readline(std.textio.input, integer_line);

    assert integer_line'length >= 0
	report "(join.vhdl) Missing integer in identifiers file."
	severity failure;

    read(integer_line, integer_id);
    id <= integer_id;

    wait for 0 fs;

    wait;

end process;


join : process

   variable number_tokens_passed : integer := 0;
   variable mesg : line;
   
  begin

--   for I in 1 to number_in loop
     in_arc <= null;
--   end loop;
     out_arc <= null;

   loop
     wait on in_arc;
     for I in 1 to number_in loop
       if in_arc(i).handshake = sending then
	if (debug_control = handshaking or 
	    debug_control = summary) then 
	print_handshake_message ("Join unit: ", name," received handshake signal: ", 
                                 in_arc(i).handshake,
                                 " from component: ", in_arc(i).id,
				 " at Time: ", now);
	end if;
        out_arc <= in_arc(i);
        number_tokens_passed := number_tokens_passed + 1;
	if (debug_control = trace_tokens or 
	    debug_control = summary) then
	write(mesg, string'("Total number of tokens passed through Join unit "));
	write(mesg, name);
	write(mesg, string'(" at time: "));
	write(mesg, now);
	write(mesg, string'(" = "));
	write(mesg, number_tokens_passed);
	writeline(outfile, mesg);
	print_joinfork_message ("Join unit: ", name, " passing tokens to ",
                                "output arc", " at Time: ", now);
	end if;
        wait on out_arc until (out_arc.handshake = accepted);
	if (debug_control = handshaking or 
	    debug_control = summary) then
	print_handshake_message ("Join unit: ", name," received handshake signal: ", 
				 out_arc.handshake,
                                 " from component: ", out_arc.id,
				 " at Time: ", now);
	end if;	
        out_arc <= null;
        wait for 0 fs;
        in_arc(i) <= (id, accepted, FALSE, now);
        wait for 0 fs;
	if (debug_control = handshaking or 
	    debug_control = summary) then
	print_handshake_message ("Join unit: ", name," sent handshake signal: ", 
				 in_arc(i).handshake,
                                 " to component: ", in_arc(i).id,
				 " at Time: ", now);	
	end if;
        in_arc(i) <= null;
        wait for 0 fs;
       end if;
     end loop;
   end loop;
 end process;

end behavior;
